/* pgen.c  parser generator
   generate a parser for LL(1) grammar

   A -> B | C #
   B -> a B | nil #

   the program can be embedded with $str

   A -> B $str . . .

				24 Apr 2001
   include local var		28 Apr 2001
   fix error when the first term (RHS) begin with $str
   generate function prototype  5 Jan 2002
*/

#include "compile.h"

#define		MAXBUF		5000
#define		GOTO		":goto:"
#define		isterm()	( *cp == 't' && *(cp+1) == 'k' )

char inbuf[MAXBUF], *cp, head[256], prev[256];
int first;

FILE *FH;			/* prototype in header file  */

readspec() {
  FILE *fin;
  fin = fopen("parse.txt","rt");
  fread(inbuf,1,MAXBUF,fin);
  fclose(fin);
}

int is(char *s) { return (strcmp(cp,s) == 0); }
int eqs(char *s1, char *s2) { return (strcmp(s1,s2)==0); }

token() { cp = strtok(NULL, " \t\n"); /* printf(">> %s\n",cp); */ }

dogen() {
  if( cp == NULL ) return;
  strcpy(head,cp);
  printf("int %s() { \n",head);
  fprintf(FH,"int %s(void);\n",head);
  token();
  if( is("[") ) { 	/* declare local var */
    token();
    while( !is("]") ) {
      printf("int %s; ",cp);
      token();
    }
    printf("\n");
    token(); 		   /* read -> */
  }
  printf(" loop: \n");
  doRHS();
}

doRHS() {
  first = 1;
  token();
  while( 1 ) {
    if( isterm() ) {
      if( first ) {
		printf("if( tok == %s ) {\n",cp);
		first = 0;
      } else {
		printf("expect(%s, \"missing %s\");\n",cp,cp);
      }
      trylex();
    } else
    if( *cp == '$' ) {
      printf("%s\n", cp+1);
      next();
    } else
    if( is("|") ) {
      first = 1;
      if( !eqs(prev,GOTO) ) printf("return 1; }\n");
      next();
    } else
    if( is("nil") ) {
      printf("return 1;\n}\n\n");
      next();
    } else
    if( is("#") ) {
      break;
    } else
    /* is non-term */
    if( eqs(head, cp) ) {
      printf("goto loop; }\n");
      strcpy(prev,GOTO);
      token();
    } else {
    	if( first ) {
	      printf("if( %s() ) {\n",cp);
	      first = 0;
	      next();
	    } else {
	      printf("commit(%s());\n",cp);
	      next();
	    }
	}
  }
  if( !eqs(prev,"nil") )
    printf("return 1; }\nreturn 0;\n}\n\n");
}

next() {  strcpy(prev,cp); token(); }

trylex() {
  next();
  if( *cp == '$' ) {
    printf("%s lex();\n",cp+1);
    next();
  }
  else printf("lex();\n");
}


main() {
  readspec();
  FH = fopen("parse.h","w");
  cp = strtok(inbuf, " \t\n");
  while( !eqs(cp,"END") ) {
    if( is("$") ) {		/* print line to output */
      cp = strtok(NULL,"\n");
      printf("%s\n",cp);
    } else
    if( is("//") ) {		/* ignore comment line */
      cp = strtok(NULL,"\n");
    } else {
      dogen();
    }
    token();
  }
  fclose(FH);
}

